import discover from "@watsonb8/wiz-lib/build/discovery";
import {
    API,
    Characteristic,
    DynamicPlatformPlugin,
    Logger,
    Logging,
    PlatformAccessory,
    PlatformConfig,
    Service,
    UnknownContext,
} from "homebridge";
import { v3 } from "node-hue-api";
import Api from "node-hue-api/lib/api/Api";
import { FluxAccessory } from "./fluxAccessory";
import { IConfig } from "./models/iConfig";
import { PLATFORM_NAME, PLUGIN_NAME } from "./settings";
import { Sleep } from "./sleep";

export class Platform implements DynamicPlatformPlugin {
    private hue: Api | undefined;
    private accessory: PlatformAccessory | undefined = undefined;
    private config: IConfig;
    constructor(
        public readonly log: Logger,
        config: PlatformConfig,
        public readonly api: API
    ) {
        this.config = config as unknown as IConfig;
        this.log.info("INFO - Registering Flux platform");
        this.api.on("didFinishLaunching", this.didFinishLaunching.bind(this));
    }

    async configureAccessory(accessory: PlatformAccessory<UnknownContext>) {
        this.log.info("Loading accessory from cache:", accessory.displayName);

        this.accessory = accessory;
    }

    /**
     * Handler for didFinishLaunching
     * Happens after constructor
     */
    private async didFinishLaunching() {
        this.log.info(`INFO - Done registering Flux platform`);

        await this.connectHue();
        const wizBulbs = await this.connectWiz();

        this.log.info("Registering accessory: " + this.config.name);
        const uuid = this.api.hap.uuid.generate(this.config.name);

        // Load accessory if not cached
        if (!this.accessory) {
            this.accessory = new this.api.platformAccessory(
                this.config.name,
                uuid
            );
            this.accessory.context["DeviceName"] = this.config.name;
            this.accessory.context["Type"] = typeof "Flux";
            this.api.registerPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [
                this.accessory,
            ]);
        }

        new FluxAccessory({
            platform: this,
            accessory: this.accessory,
            hue: this.hue!,
            wizBulbs: wizBulbs ?? [],
            config: this.config,
        });
    }

    private connectWiz = async () => {
        if (!this.config) {
            return;
        }

        return await discover();
    };

    private connectHue = async () => {
        if (!this.config) {
            return;
        }

        if (this.config.userName && this.config.clientKey) {
            this.hue = await v3.api
                .createLocal(this.config.ipAddress)
                .connect(
                    this.config.userName,
                    this.config.clientKey,
                    undefined
                );
            this.log.info("Using existing connection info");
        } else {
            const unauthenticatedApi = await v3.api
                .createLocal(this.config.ipAddress)
                .connect(undefined, undefined, undefined);
            let createdUser;
            let connected = false;
            while (!connected) {
                try {
                    this.log.info("Creating hue user. Push link button");
                    createdUser = await unauthenticatedApi.users.createUser(
                        "homebridge",
                        "HueChase"
                    );

                    this.hue = await v3.api
                        .createLocal(this.config.ipAddress)
                        .connect(
                            createdUser.username,
                            createdUser.clientKey,
                            undefined
                        );
                    this.log.info("Connected to Hue Bridge");
                    this.log.info(
                        `UserName: ${createdUser.username}, ClientKey: ${createdUser.clientkey}`
                    );
                    connected = true;
                } catch (err: any) {
                    if (err.getHueErrorType() === 101) {
                        this.log.info(
                            "The Link button on the bridge was not pressed. Please press the Link button and try again."
                        );
                        Sleep(5000);
                    } else {
                        this.log.info(`Unexpected Error: ${err.message}`);
                        break;
                    }
                }
            }
        }
    };
}
